-- one time operations on ubs public schema

-- add last_seen to device
ALTER TABLE device ADD COLUMN IF NOT EXISTS last_seen TIMESTAMPTZ DEFAULT NULL;

-- cust_device_link_1 ---
CREATE TABLE IF NOT EXISTS cust_device_link_1 (
  id BIGSERIAL PRIMARY KEY NOT NULL,
  cust_ix_link_id BIGINT NOT NULL REFERENCES cust_ix_link(id),
  device_id BIGINT NOT NULL REFERENCES device(id),
  first_seen TIMESTAMPTZ NOT NULL
);

CREATE UNIQUE INDEX IF NOT EXISTS idx_cust_device_link_1_cust_ix_link_id_device_id ON cust_device_link_1(cust_ix_link_id, device_id);

-- cust_device_link_3 ---
CREATE TABLE IF NOT EXISTS cust_device_link_3 (
  id BIGSERIAL PRIMARY KEY NOT NULL,
  cust_ix_link_id BIGINT NOT NULL REFERENCES cust_ix_link(id),
  device_id BIGINT NOT NULL REFERENCES device(id),
  first_seen TIMESTAMPTZ NOT NULL
);

CREATE UNIQUE INDEX IF NOT EXISTS idx_cust_device_link_3_cust_ix_link_id_device_id ON cust_device_link_3(cust_ix_link_id, device_id);

-- ubs_table_rotation
CREATE TABLE IF NOT EXISTS ubs_table_rotation (
    id BIGSERIAL PRIMARY KEY NOT NULL,
    table_name VARCHAR,
    previous VARCHAR,
    current VARCHAR,
    next VARCHAR
);

CREATE UNIQUE INDEX IF NOT EXISTS idx_ubs_table_rotation_table_name ON ubs_table_rotation(table_name);

INSERT INTO ubs_table_rotation(table_name, previous, current, next)
VALUES ('cust_device_link', 'cust_device_link_1', 'cust_device_link', 'cust_device_link_3')
ON CONFLICT (table_name) DO NOTHING ;

-- TPLAT-11882 rotate device tables for Mosaic2 --
WITH existing(pre, curr, next) AS (
    SELECT previous, current, next
FROM ubs_table_rotation WHERE table_name='cust_device_link'
    )
UPDATE ubs_table_rotation
SET previous=e.curr, current=e.next, next=e.pre
    FROM existing AS e
WHERE table_name='cust_device_link' AND e.curr = 'cust_device_link';

-- store functions --

DO $$ BEGIN
    CREATE TYPE device_insert_v3 AS (sha256 bytea, org_key VARCHAR(128), device_id BIGINT, device_name VARCHAR(256),
                              first_seen TIMESTAMPTZ);
EXCEPTION
    WHEN duplicate_object THEN null;
END $$;

CREATE OR REPLACE FUNCTION insert_device_v3(insertion_data device_insert_v3[])
  RETURNS bytea[] AS $$
  DECLARE ret bytea[];
  BEGIN
    INSERT INTO device (device_id, device_name)
      (SELECT device_id, device_name FROM unnest(insertion_data))
      ON CONFLICT DO NOTHING;

    -- MATERIALIZED needs to be here or Postgresql will "optimize"
    -- the data1 query with the rest making it very un-optimized and slow.
    -- https://www.postgresql.org/docs/current/queries-with.html#QUERIES-WITH-CTE-MATERIALIZATION
    WITH data1(c_id, fi_id, d_id, sha256, first_seen) AS MATERIALIZED (
    SELECT c.id, fi.id, d.id, i.sha256, i.first_seen
       FROM unnest(insertion_data) i
       JOIN device d ON i.device_id = d.device_id AND i.device_name = d.device_name
       JOIN cust c ON digest(i.org_key, 'sha256') = c.search_key
       JOIN file_index fi ON i.sha256 = fi.sha256
    ), data2(cixl_id, d_id, sha256, first_seen) AS (
    SELECT cixl.id, d_id, sha256, first_seen
        FROM data1 d
        JOIN cust_ix_link cixl ON c_id = cixl.cust_id AND d.fi_id = cixl.file_index_id
    ), ins (cixl_id, d_id) AS (
      INSERT INTO cust_device_link(cust_ix_link_id, device_id, first_seen)
      SELECT cixl_id, d_id, first_seen
      FROM data2
      ON CONFLICT (cust_ix_link_id, device_id) DO NOTHING
      RETURNING cust_ix_link_id, device_id
    )
    SELECT array(
      SELECT data2.sha256
      FROM data2
      EXCEPT (
        SELECT data2.sha256
        FROM data2
        JOIN ins ON data2.cixl_id = ins.cixl_id AND data2.d_id = ins.d_id
      )
    )
    INTO ret;
    RETURN ret;

  END;
$$ LANGUAGE plpgsql;


-- write to device and cust_device_link_1
CREATE OR REPLACE FUNCTION insert_device_1_v3(insertion_data device_insert_v3[])
  RETURNS bytea[] AS $$
  DECLARE ret bytea[];
  BEGIN
    INSERT INTO device (device_id, device_name)
      (SELECT device_id, device_name FROM unnest(insertion_data))
      ON CONFLICT DO NOTHING;

    WITH data1(c_id, fi_id, d_id, sha256, first_seen) AS MATERIALIZED (
    SELECT c.id, fi.id, d.id, i.sha256, i.first_seen
       FROM unnest(insertion_data) i
       JOIN device d ON i.device_id = d.device_id AND i.device_name = d.device_name
       JOIN cust c ON digest(i.org_key, 'sha256') = c.search_key
       JOIN file_index fi ON i.sha256 = fi.sha256
    ), data2(cixl_id, d_id, sha256, first_seen) AS (
    SELECT cixl.id, d_id, sha256, first_seen
        FROM data1 d
        JOIN cust_ix_link cixl ON c_id = cixl.cust_id AND d.fi_id = cixl.file_index_id
    ), ins (cixl_id, d_id) AS (
      INSERT INTO cust_device_link_1(cust_ix_link_id, device_id, first_seen)
      SELECT cixl_id, d_id, first_seen
      FROM data2
      ON CONFLICT (cust_ix_link_id, device_id) DO NOTHING
      RETURNING cust_ix_link_id, device_id
    )
    SELECT array(
      SELECT data2.sha256
      FROM data2
      EXCEPT (
        SELECT data2.sha256
        FROM data2
        JOIN ins ON data2.cixl_id = ins.cixl_id AND data2.d_id = ins.d_id
      )
    )
    INTO ret;
    RETURN ret;

  END;
$$ LANGUAGE plpgsql;


-- write to device and cust_device_link_3 as rotation/cache table
CREATE OR REPLACE FUNCTION insert_device_3_v3(insertion_data device_insert_v3[])
  RETURNS bytea[] AS $$
  DECLARE ret bytea[];
  BEGIN
    INSERT INTO device (device_id, device_name)
      (SELECT device_id, device_name FROM unnest(insertion_data))
      ON CONFLICT DO NOTHING;

    WITH data1(c_id, fi_id, d_id, sha256, first_seen) AS MATERIALIZED (
    SELECT c.id, fi.id, d.id, i.sha256, i.first_seen
       FROM unnest(insertion_data) i
       JOIN device d ON i.device_id = d.device_id AND i.device_name = d.device_name
       JOIN cust c ON digest(i.org_key, 'sha256') = c.search_key
       JOIN file_index fi ON i.sha256 = fi.sha256
    ), data2(cixl_id, d_id, sha256, first_seen) AS (
    SELECT cixl.id, d_id, sha256, first_seen
        FROM data1 d
        JOIN cust_ix_link cixl ON c_id = cixl.cust_id AND d.fi_id = cixl.file_index_id
    ), ins (cixl_id, d_id) AS (
      INSERT INTO cust_device_link_3(cust_ix_link_id, device_id, first_seen)
      SELECT cixl_id, d_id, first_seen
      FROM data2
      ON CONFLICT (cust_ix_link_id, device_id) DO NOTHING
      RETURNING cust_ix_link_id, device_id
    )
    SELECT array(
      SELECT data2.sha256
      FROM data2
      EXCEPT (
        SELECT data2.sha256
        FROM data2
        JOIN ins ON data2.cixl_id = ins.cixl_id AND data2.d_id = ins.d_id
      )
    )
    INTO ret;
    RETURN ret;

  END;
$$ LANGUAGE plpgsql;


-- check cust_device_link_1, trade redundency for performance
CREATE OR REPLACE FUNCTION select_duplicate_cust_device_links_1_v3(insertion_data hash_cust_device_v3[])
    RETURNS TABLE (sha256 varchar, org_key varchar, device_id bigint, device_name varchar) AS $$
BEGIN
    RETURN QUERY
    WITH data1(c_id, fi_id, d_id, sha256, org_key, device_id, device_name) AS (
      SELECT c.id, fi.id, d.id, i.sha256, i.org_key, i.device_id, i.device_name
      FROM unnest(insertion_data) i
      JOIN device d ON i.device_id = d.device_id AND i.device_name = d.device_name
      JOIN cust c ON digest(i.org_key, 'sha256') = c.search_key
      JOIN file_index fi ON decode(i.sha256, 'hex') = fi.sha256
    ), data2(cixl_id, d_id, sha256, org_key, device_id, device_name) AS (
      SELECT cixl.id, d.d_id, d.sha256, d.org_key, d.device_id, d.device_name
      FROM data1 d
      JOIN cust_ix_link cixl ON d.c_id = cixl.cust_id AND d.fi_id = cixl.file_index_id
    )
    SELECT DISTINCT
      d.sha256
      , d.org_key
      , d.device_id
      , d.device_name
    FROM data2 d
    JOIN cust_device_link_1 cdl ON d.cixl_id = cdl.cust_ix_link_id AND d.d_id = cdl.device_id;
END;
$$ LANGUAGE plpgsql;

-- check cust_device_link_3
CREATE OR REPLACE FUNCTION select_duplicate_cust_device_links_3_v3(insertion_data hash_cust_device_v3[])
    RETURNS TABLE (sha256 varchar, org_key varchar, device_id bigint, device_name varchar) AS $$
BEGIN
    RETURN QUERY
    WITH data1(c_id, fi_id, d_id, sha256, org_key, device_id, device_name) AS (
      SELECT c.id, fi.id, d.id, i.sha256, i.org_key, i.device_id, i.device_name
      FROM unnest(insertion_data) i
      JOIN device d ON i.device_id = d.device_id AND i.device_name = d.device_name
      JOIN cust c ON digest(i.org_key, 'sha256') = c.search_key
      JOIN file_index fi ON decode(i.sha256, 'hex') = fi.sha256
    ), data2(cixl_id, d_id, sha256, org_key, device_id, device_name) AS (
      SELECT cixl.id, d.d_id, d.sha256, d.org_key, d.device_id, d.device_name
      FROM data1 d
      JOIN cust_ix_link cixl ON d.c_id = cixl.cust_id AND d.fi_id = cixl.file_index_id
    )
    SELECT DISTINCT
      d.sha256
      , d.org_key
      , d.device_id
      , d.device_name
    FROM data2 d
    JOIN cust_device_link_3 cdl ON d.cixl_id = cdl.cust_ix_link_id AND d.d_id = cdl.device_id;
END;
$$ LANGUAGE plpgsql;